import copy
import dataclasses
import json
from pathlib import Path

import pytest

from randovania.gui import tracker_window
from randovania.layout.lib.teleporters import TeleporterShuffleMode
from randovania.layout.preset_migration import VersionedPreset
from randovania.layout.prime2.translator_configuration import LayoutTranslatorRequirement


@pytest.fixture(params=[{},
                        {"elevators": TeleporterShuffleMode.ONE_WAY_ANYTHING,
                         "translator_configuration": True}],
                name="layout_config")
def _layout_config(request, default_layout_configuration):
    if "translator_configuration" in request.param:
        translator_requirement = copy.copy(default_layout_configuration.translator_configuration.translator_requirement)
        for gate in translator_requirement.keys():
            translator_requirement[gate] = LayoutTranslatorRequirement.RANDOM
            break

        new_gate = dataclasses.replace(default_layout_configuration.translator_configuration,
                                       translator_requirement=translator_requirement)
        request.param["translator_configuration"] = new_gate
    return dataclasses.replace(default_layout_configuration, **request.param)


def test_load_previous_state_no_previous_layout(tmp_path: Path, default_layout_configuration):
    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_previous_layout_not_json(tmp_path: Path, default_layout_configuration):
    # Setup
    tmp_path.joinpath("preset.rdvpreset").write_text("this is not a json")

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_previous_layout_not_layout(tmp_path: Path, default_layout_configuration):
    # Setup
    tmp_path.joinpath("preset.rdvpreset").write_text(json.dumps({"trick_level": "foo"}))
    tmp_path.joinpath("state.json").write_text("[]")

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_layout_configuration)

    # Assert
    assert result is None


def test_load_previous_state_missing_state(tmp_path: Path, default_preset):
    # Setup
    VersionedPreset.with_preset(default_preset).save_to_file(tmp_path.joinpath("preset.rdvpreset"))

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_preset.configuration)

    # Assert
    assert result is None


def test_load_previous_state_invalid_state(tmp_path: Path, default_preset):
    # Setup
    VersionedPreset.with_preset(default_preset).save_to_file(tmp_path.joinpath("preset.rdvpreset"))
    tmp_path.joinpath("state.json").write_text("")

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_preset.configuration)

    # Assert
    assert result is None


def test_load_previous_state_success(tmp_path: Path, default_preset):
    # Setup
    data = {"asdf": 5, "zxcv": 123}
    VersionedPreset.with_preset(default_preset).save_to_file(tmp_path.joinpath("preset.rdvpreset"))
    tmp_path.joinpath("state.json").write_text(json.dumps(data))

    # Run
    result = tracker_window._load_previous_state(tmp_path, default_preset.configuration)

    # Assert
    assert result == data


@pytest.mark.parametrize("shuffle_advanced", [False, True])
def test_apply_previous_state(skip_qtbot, tmp_path: Path, default_preset, shuffle_advanced):
    if shuffle_advanced:
        translator_requirement = copy.copy(
            default_preset.configuration.translator_configuration.translator_requirement)
        for gate in translator_requirement.keys():
            translator_requirement[gate] = LayoutTranslatorRequirement.RANDOM
            break

        new_gate = dataclasses.replace(default_preset.configuration.translator_configuration,
                                       translator_requirement=translator_requirement)
        layout_config = dataclasses.replace(
            default_preset.configuration,
            elevators=dataclasses.replace(
                default_preset.configuration.elevators,
                mode=TeleporterShuffleMode.ONE_WAY_ANYTHING,
            ),
            translator_configuration=new_gate)

        preset = dataclasses.replace(default_preset.fork(), configuration=layout_config)

    else:
        preset = default_preset

    state = {
        "actions": [
            0,
            4
        ],
        "collected_pickups": {
            'Amber Translator': 0,
            'Annihilator Beam': 0,
            'Boost Ball': 0,
            'Cobalt Translator': 0,
            'Dark Agon Key 1': 0,
            'Dark Agon Key 2': 0,
            'Dark Agon Key 3': 0,
            'Dark Ammo Expansion': 0,
            'Dark Beam': 0,
            'Dark Torvus Key 1': 0,
            'Dark Torvus Key 2': 0,
            'Dark Torvus Key 3': 0,
            'Dark Visor': 0,
            'Darkburst': 0,
            'Echo Visor': 0,
            'Emerald Translator': 0,
            'Energy Tank': 0,
            'Grapple Beam': 0,
            'Gravity Boost': 0,
            'Ing Hive Key 1': 0,
            'Ing Hive Key 2': 0,
            'Ing Hive Key 3': 0,
            'Light Ammo Expansion': 0,
            'Light Beam': 0,
            'Missile Expansion': 0,
            'Missile Launcher': 0,
            'Morph Ball Bomb': 0,
            'Power Bomb': 0,
            'Power Bomb Expansion': 0,
            'Progressive Suit': 0,
            'Screw Attack': 0,
            'Seeker Launcher': 0,
            'Sky Temple Key 1': 0,
            'Sky Temple Key 2': 0,
            'Sky Temple Key 3': 0,
            'Sky Temple Key 4': 0,
            'Sky Temple Key 5': 0,
            'Sky Temple Key 6': 0,
            'Sky Temple Key 7': 0,
            'Sky Temple Key 8': 0,
            'Sky Temple Key 9': 0,
            'Sonic Boom': 0,
            'Space Jump Boots': 1,
            'Spider Ball': 0,
            'Sunburst': 0,
            'Super Missile': 0,
            'Violet Translator': 0,
        },
        "elevators": [
            {'data': None,
             'teleporter': {'area_asset_id': 1473133138,
                            'instance_id': 122,
                            'world_asset_id': 1119434212}},
            {'data': None,
             'teleporter': {'area_asset_id': 2806956034,
                            'instance_id': 1245307,
                            'world_asset_id': 1119434212}},
            {'data': None,
             'teleporter': {'area_asset_id': 3331021649,
                            'instance_id': 2949235,
                            'world_asset_id': 1119434212}},
            {'data': None,
             'teleporter': {'area_asset_id': 2556480432,
                            'instance_id': 393260,
                            'world_asset_id': 2252328306}},
            {'data': None,
             'teleporter': {'area_asset_id': 408633584,
                            'instance_id': 152,
                            'world_asset_id': 2252328306}},
            {'data': None,
             'teleporter': {'area_asset_id': 2399252740,
                            'instance_id': 524321,
                            'world_asset_id': 2252328306}},
            {'data': None,
             'teleporter': {'area_asset_id': 3528156989,
                            'instance_id': 38,
                            'world_asset_id': 464164546}},
            {'data': None,
             'teleporter': {'area_asset_id': 900285955,
                            'instance_id': 1245332,
                            'world_asset_id': 464164546}},
            {'data': None,
             'teleporter': {'area_asset_id': 3145160350,
                            'instance_id': 1638535,
                            'world_asset_id': 464164546}},
            {'data': None,
             'teleporter': {'area_asset_id': 1660916974,
                            'instance_id': 1572998,
                            'world_asset_id': 1006255871}},
            {'data': None,
             'teleporter': {'area_asset_id': 1287880522,
                            'instance_id': 2097251,
                            'world_asset_id': 1006255871}},
            {'data': None,
             'teleporter': {'area_asset_id': 3455543403,
                            'instance_id': 3342446,
                            'world_asset_id': 1006255871}},
            {'data': None,
             'teleporter': {'area_asset_id': 1345979968,
                            'instance_id': 3538975,
                            'world_asset_id': 1006255871}},
            {'data': None,
             'teleporter': {'area_asset_id': 2889020216,
                            'instance_id': 1966093,
                            'world_asset_id': 1006255871}},
            {'data': None,
             'teleporter': {'area_asset_id': 2918020398,
                            'instance_id': 589851,
                            'world_asset_id': 1006255871}},
            {'data': None,
             'teleporter': {'area_asset_id': 3205424168,
                            'instance_id': 4522032,
                            'world_asset_id': 1039999561}},
            {'data': None,
             'teleporter': {'area_asset_id': 1868895730,
                            'instance_id': 129,
                            'world_asset_id': 1039999561}},
            {'data': None,
             'teleporter': {'area_asset_id': 3479543630,
                            'instance_id': 2162826,
                            'world_asset_id': 1039999561}}
        ],
        "translator_gates": {
            "0": None,
            '1': None,
            '10': None,
            '11': None,
            '12': None,
            '13': None,
            '14': None,
            '15': None,
            '16': None,
            '2': None,
            '3': None,
            '4': None,
            '5': None,
            '6': None,
            '7': None,
            '8': None,
            '9': None,
        },
        "starting_location": {
            "world_asset_id": 1006255871,
            "area_asset_id": 1655756413
        }
    }
    if shuffle_advanced:
        for elevator in state["elevators"]:
            if elevator["teleporter"]["instance_id"] == 2949235:
                elevator["data"] = {'area_asset_id': 50083607, 'world_asset_id': 1119434212}
        state["translator_gates"]['0'] = 97
    VersionedPreset.with_preset(preset).save_to_file(tmp_path.joinpath("preset.rdvpreset"))
    tmp_path.joinpath("state.json").write_text(json.dumps(state), "utf-8")

    # Run
    window = tracker_window.TrackerWindow(tmp_path, preset)
    skip_qtbot.add_widget(window)

    # Assert
    assert window.state_for_current_configuration() is not None
    persisted_data = json.loads(tmp_path.joinpath("state.json").read_text("utf-8"))
    assert persisted_data == state

    window.reset()
    window.persist_current_state()

    persisted_data = json.loads(tmp_path.joinpath("state.json").read_text("utf-8"))
    assert persisted_data != state



import dataclasses
import functools

from PySide2 import QtWidgets, QtCore

from randovania.game_description.game_description import GameDescription
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.gui.dialog.trick_details_popup import TrickDetailsPopup, ResourceDetailsPopup
from randovania.gui.generated.preset_trick_level_ui import Ui_PresetTrickLevel
from randovania.gui.lib import trick_lib, signal_handling
from randovania.gui.lib.window_manager import WindowManager
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.base.trick_level import LayoutTrickLevel
from randovania.layout.base.logical_resource_action import LayoutLogicalResourceAction
from randovania.layout.preset import Preset
from randovania.lib import enum_lib
from randovania.gui.lib import common_qt_lib

_GAME_SPECIFIC_MINIMAL_LOGIC_TEXT = {
    RandovaniaGame.METROID_PRIME: (
        "TODO: Minimal logic description for Prime."
    ),
    RandovaniaGame.METROID_PRIME_ECHOES: (
        "Randovania will only check that Screw Attack, Dark Visor and Light Suit won't all be behind "
        "Ing Caches and Dark Water, removing the biggest reasons for a pure random layout to be impossible."
    ),
    RandovaniaGame.METROID_PRIME_CORRUPTION: (
        "TODO: Minimal logic description for Corruption."
    ),
}


class PresetTrickLevel(PresetTab, Ui_PresetTrickLevel):

    def __init__(self, editor: PresetEditor, game_description: GameDescription, window_manager: WindowManager):
        super().__init__(editor)
        self.setupUi(self)

        self.game_description = game_description
        self._window_manager = window_manager

        self.dangerous_combo.setItemData(0, LayoutLogicalResourceAction.RANDOMLY)
        self.dangerous_combo.setItemData(1, LayoutLogicalResourceAction.LAST_RESORT)
        signal_handling.on_combo(self.dangerous_combo, self._on_dangerous_changed)

        self.trick_level_layout.setAlignment(QtCore.Qt.AlignTop)
        signal_handling.on_checked(self.trick_level_minimal_logic_check, self._on_trick_level_minimal_logic_check)
        signal_handling.on_checked(self.underwater_abuse_check, self._on_underwater_abuse_check)
        self.underwater_abuse_label.linkActivated.connect(self._on_click_link_underwater_details)

        self.trick_difficulties_layout = QtWidgets.QGridLayout()
        self._slider_for_trick = {}

        tricks_in_use = trick_lib.used_tricks(self.game_description)
        size_policy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)

        self.trick_level_line_1.setVisible(self.game_enum != RandovaniaGame.METROID_PRIME_CORRUPTION)
        self.underwater_abuse_label.setText(self.underwater_abuse_label.text().replace("color:#0000ff;", ""))

        for w in [self.trick_level_minimal_logic_check, self.trick_level_minimal_logic_label]:
            w.setVisible(self.game_enum in _GAME_SPECIFIC_MINIMAL_LOGIC_TEXT)
        self.trick_level_minimal_logic_label.setText(
            self.trick_level_minimal_logic_label.text().format(
                game_specific_text=_GAME_SPECIFIC_MINIMAL_LOGIC_TEXT.get(self.game_enum, "Unknown text")
            )
        )

        if self.game_enum != RandovaniaGame.METROID_PRIME:
            for w in [self.underwater_abuse_check, self.underwater_abuse_label]:
                w.setVisible(False)

        self._create_difficulty_details_row()

        row = 2
        for trick in sorted(self.game_description.resource_database.trick, key=lambda _trick: _trick.long_name):
            if trick not in tricks_in_use:
                continue

            if row > 1:
                self.trick_difficulties_layout.addItem(QtWidgets.QSpacerItem(20, 40,
                                                                             QtWidgets.QSizePolicy.Minimum,
                                                                             QtWidgets.QSizePolicy.Expanding))

            trick_label = QtWidgets.QLabel(self.trick_level_scroll_contents)
            trick_label.setSizePolicy(size_policy)
            trick_label.setWordWrap(True)
            trick_label.setFixedWidth(100)
            trick_label.setText(trick.long_name)

            self.trick_difficulties_layout.addWidget(trick_label, row, 1, 1, 1)

            slider_layout = QtWidgets.QGridLayout()
            slider_layout.setHorizontalSpacing(0)
            for i in range(12):
                slider_layout.setColumnStretch(i, 1)

            horizontal_slider = QtWidgets.QSlider(self.trick_level_scroll_contents)
            horizontal_slider.setMaximum(5)
            horizontal_slider.setPageStep(2)
            horizontal_slider.setOrientation(QtCore.Qt.Horizontal)
            horizontal_slider.setTickPosition(QtWidgets.QSlider.TicksBelow)
            horizontal_slider.setEnabled(False)
            horizontal_slider.valueChanged.connect(functools.partial(self._on_slide_trick_slider, trick))
            self._slider_for_trick[trick] = horizontal_slider
            slider_layout.addWidget(horizontal_slider, 0, 1, 1, 10)

            used_difficulties = trick_lib.difficulties_for_trick(self.game_description, trick)
            for i, trick_level in enumerate(enum_lib.iterate_enum(LayoutTrickLevel)):
                if trick_level == LayoutTrickLevel.DISABLED or trick_level in used_difficulties:
                    difficulty_label = QtWidgets.QLabel(self.trick_level_scroll_contents)
                    difficulty_label.setAlignment(QtCore.Qt.AlignHCenter)
                    difficulty_label.setText(trick_level.long_name)

                    slider_layout.addWidget(difficulty_label, 1, 2 * i, 1, 2)

            self.trick_difficulties_layout.addLayout(slider_layout, row, 2, 1, 1)

            if self._window_manager is not None:
                tool_button = QtWidgets.QToolButton(self.trick_level_scroll_contents)
                tool_button.setText("?")
                tool_button.clicked.connect(functools.partial(self._open_trick_details_popup, trick))
                self.trick_difficulties_layout.addWidget(tool_button, row, 3, 1, 1)

            row += 1

        self.trick_level_layout.addLayout(self.trick_difficulties_layout)

    @property
    def uses_patches_tab(self) -> bool:
        return False

    @property
    def game_enum(self):
        return self.game_description.game

    def on_preset_changed(self, preset: Preset):
        trick_level_configuration = preset.configuration.trick_level
        self.trick_level_minimal_logic_check.setChecked(trick_level_configuration.minimal_logic)
        common_qt_lib.set_combo_with_value(self.dangerous_combo, preset.configuration.logical_resource_action)

        for trick, slider in self._slider_for_trick.items():
            assert self._slider_for_trick[trick] is slider
            slider.setValue(trick_level_configuration.level_for_trick(trick).as_number)
            slider.setEnabled(not trick_level_configuration.minimal_logic)

        if self.game_enum == RandovaniaGame.METROID_PRIME:
            self.underwater_abuse_check.setChecked(preset.configuration.allow_underwater_movement_without_gravity)

    def _create_difficulty_details_row(self):
        row = 1

        trick_label = QtWidgets.QLabel(self.trick_level_scroll_contents)
        trick_label.setWordWrap(True)
        trick_label.setText("Difficulty Details")

        self.trick_difficulties_layout.addWidget(trick_label, row, 1, 1, -1)

        slider_layout = QtWidgets.QGridLayout()
        slider_layout.setHorizontalSpacing(0)
        for i in range(12):
            slider_layout.setColumnStretch(i, 1)
        #
        # if self._window_manager is not None:
        #     for i, trick_level in enumerate(LayoutTrickLevel):
        #         if trick_level not in {LayoutTrickLevel.NO_TRICKS, LayoutTrickLevel.MINIMAL_LOGIC}:
        #             tool_button = QtWidgets.QToolButton(self.trick_level_scroll_contents)
        #             tool_button.setText(trick_level.long_name)
        #             tool_button.clicked.connect(functools.partial(self._open_difficulty_details_popup, trick_level))
        #
        #             slider_layout.addWidget(tool_button, 1, 2 * i, 1, 2)

        self.trick_difficulties_layout.addLayout(slider_layout, row, 2, 1, 1)

    def _on_slide_trick_slider(self, trick: TrickResourceInfo, value: int):
        if self._slider_for_trick[trick].isEnabled():
            with self._editor as options:
                options.set_configuration_field(
                    "trick_level",
                    options.configuration.trick_level.set_level_for_trick(
                        trick,
                        LayoutTrickLevel.from_number(value)
                    )
                )

    def _on_trick_level_minimal_logic_check(self, state: bool):
        with self._editor as options:
            options.set_configuration_field(
                "trick_level",
                dataclasses.replace(options.configuration.trick_level,
                                    minimal_logic=state)
            )

    def _on_underwater_abuse_check(self, state: bool):
        with self._editor as options:
            options.set_configuration_field(
                "allow_underwater_movement_without_gravity",
                state,
            )

    def _on_dangerous_changed(self, value: LayoutLogicalResourceAction):
        with self._editor as editor:
            editor.set_configuration_field("logical_resource_action", value)

    def _on_click_link_underwater_details(self, link: str):
        self._exec_trick_details(ResourceDetailsPopup(
            self,
            self._window_manager,
            self.game_description,
            self.game_description.resource_database.get_by_type_and_index(ResourceType.MISC, 0),
        ))

    def _exec_trick_details(self, popup: TrickDetailsPopup):
        self._trick_details_popup = popup
        self._trick_details_popup.setWindowModality(QtCore.Qt.WindowModal)
        self._trick_details_popup.open()

    def _open_trick_details_popup(self, trick: TrickResourceInfo):
        self._exec_trick_details(TrickDetailsPopup(
            self,
            self._window_manager,
            self.game_description,
            trick,
            self._editor.configuration.trick_level.level_for_trick(trick),
        ))



import collections
from typing import List, Iterable, Tuple, Dict

from randovania.game_description import default_database
from randovania.game_description.item.major_item import MajorItem
from randovania.games.game import RandovaniaGame
from randovania.generator.item_pool import pool_creator
from randovania.layout.base.ammo_configuration import AmmoConfiguration
from randovania.layout.base.available_locations import RandomizationMode
from randovania.layout.base.base_configuration import BaseConfiguration
from randovania.layout.base.major_item_state import MajorItemState
from randovania.layout.base.major_items_configuration import MajorItemsConfiguration
from randovania.layout.base.pickup_model import PickupModelStyle
from randovania.layout.preset import Preset
from randovania.layout.prime1.prime_configuration import PrimeConfiguration, LayoutCutsceneMode
from randovania.layout.prime2.echoes_configuration import LayoutSkyTempleKeyMode, EchoesConfiguration
from randovania.layout.prime3.corruption_configuration import CorruptionConfiguration


def _bool_to_str(b: bool) -> str:
    if b:
        return "Yes"
    else:
        return "No"


_EXPECTED_ITEMS = {
    RandovaniaGame.METROID_PRIME: {
        "Combat Visor",
        "Scan Visor",
        "Power Beam"
    },
    RandovaniaGame.METROID_PRIME_ECHOES: {
        "Combat Visor",
        "Scan Visor",
        "Morph Ball",
        "Power Beam",
        "Charge Beam",
    },
    RandovaniaGame.METROID_PRIME_CORRUPTION: {
        "Scan Visor",
        "Morph Ball",
        "Morph Ball Bombs",
        "Power Beam",
        "Charge Beam",
        "Space Jump Boots",
    },
    RandovaniaGame.SUPER_METROID: set(),
}
_CUSTOM_ITEMS = {
    "Cannon Ball",
    "Double Damage",
    "Unlimited Beam Ammo",
    "Unlimited Missiles",
}

PresetDescription = Tuple[str, List[str]]


def _require_majors_check(ammo_configuration: AmmoConfiguration, ammo_names: List[str]) -> List[bool]:
    result = [False] * len(ammo_names)

    name_index_mapping = {name: i for i, name in enumerate(ammo_names)}

    for ammo, state in ammo_configuration.items_state.items():
        if ammo.name in name_index_mapping:
            result[name_index_mapping[ammo.name]] = state.requires_major_item

    return result


def message_for_required_mains(ammo_configuration: AmmoConfiguration, message_to_item: Dict[str, str]):
    item_names = [item for item in message_to_item.values()]
    main_required = _require_majors_check(ammo_configuration, item_names)
    return dict(zip(message_to_item.keys(), main_required))


def has_shuffled_item(configuration: MajorItemsConfiguration, item_name: str) -> bool:
    for item, state in configuration.items_state.items():
        if item.name == item_name:
            return state.num_shuffled_pickups > 0
    return False


def _calculate_starting_items(game: RandovaniaGame, items_state: Dict[MajorItem, MajorItemState]) -> List[str]:
    expected_items = _EXPECTED_ITEMS[game]
    starting_items = []

    for major_item, item_state in items_state.items():
        if major_item.required:
            continue

        count = item_state.num_included_in_starting_items
        if count > 0:
            if major_item.name in expected_items:
                continue
            if count > 1:
                starting_items.append(f"{count} {major_item.name}")
            else:
                starting_items.append(major_item.name)

        elif major_item.name in expected_items:
            starting_items.append(f"No {major_item.name}")

    if starting_items:
        return starting_items
    else:
        # If an expected item is missing, it's added as "No X". So empty starting_items means it's precisely vanilla
        return ["Vanilla"]


def _calculate_item_pool(game: RandovaniaGame, configuration: MajorItemsConfiguration) -> List[str]:
    item_pool = []

    unexpected_items = _EXPECTED_ITEMS[game] | _CUSTOM_ITEMS
    if game == RandovaniaGame.METROID_PRIME_ECHOES:
        if has_shuffled_item(configuration, "Progressive Grapple"):
            unexpected_items.add("Grapple Beam")
            unexpected_items.add("Screw Attack")
        else:
            unexpected_items.add("Progressive Grapple")

        if has_shuffled_item(configuration, "Progressive Suit"):
            unexpected_items.add("Dark Suit")
            unexpected_items.add("Light Suit")
        else:
            unexpected_items.add("Progressive Suit")

    elif game == RandovaniaGame.METROID_PRIME_CORRUPTION:
        if has_shuffled_item(configuration, "Progressive Beam"):
            unexpected_items.add("Plasma Beam")
            unexpected_items.add("Nova Beam")
        else:
            unexpected_items.add("Progressive Beam")

        if has_shuffled_item(configuration, "Progressive Missile"):
            unexpected_items.add("Ice Missile")
            unexpected_items.add("Seeker Missile")
        else:
            unexpected_items.add("Progressive Missile")

    for major_item, item_state in configuration.items_state.items():
        if major_item.required:
            continue

        item_was_expected = major_item.name not in unexpected_items

        if item_state.num_shuffled_pickups > 0 or item_state.include_copy_in_original_location:
            item_in_pool = True
        else:
            item_in_pool = False

        if item_in_pool:
            if not item_was_expected:
                item_pool.append(major_item.name)
        else:
            if item_was_expected and item_state.num_included_in_starting_items == 0:
                item_pool.append(f"No {major_item.name}")

    return item_pool


def _format_params_base(configuration: BaseConfiguration,
                        ) -> Dict[str, List[str]]:
    game_description = default_database.game_description_for(configuration.game)
    major_items = configuration.major_items_configuration

    template_strings = collections.defaultdict(list)

    # Item Placement
    randomization_mode = configuration.available_locations.randomization_mode

    if major_items.minimum_random_starting_items == major_items.maximum_random_starting_items:
        random_starting_items = "{}".format(major_items.minimum_random_starting_items)
    else:
        random_starting_items = "{} to {}".format(
            major_items.minimum_random_starting_items,
            major_items.maximum_random_starting_items,
        )

    template_strings["Item Placement"].append(configuration.trick_level.pretty_description)

    if randomization_mode != RandomizationMode.FULL:
        template_strings["Item Placement"].append(f"Randomization Mode: {randomization_mode.value}")

    template_strings["Item Placement"].append(f"Dangerous Actions: {configuration.logical_resource_action.long_name}")

    if random_starting_items != "0":
        template_strings["Item Placement"].append(f"Random Starting Items: {random_starting_items}")

    # Starting Items
    template_strings["Starting Items"] = _calculate_starting_items(configuration.game, major_items.items_state)

    # Item Pool
    item_pool = _calculate_item_pool(configuration.game, major_items)

    template_strings["Item Pool"].append(
        "Size: {} of {}".format(*pool_creator.calculate_pool_item_count(configuration))
    )
    if item_pool:
        template_strings["Item Pool"].append(", ".join(item_pool))

    # Difficulty
    template_strings["Difficulty"].append(
        f"Damage Strictness: {configuration.damage_strictness.long_name}"
    )
    if configuration.pickup_model_style != PickupModelStyle.ALL_VISIBLE:
        template_strings["Difficulty"].append(f"Pickup: {configuration.pickup_model_style.long_name} "
                                              f"({configuration.pickup_model_data_source.long_name})")

    # Gameplay
    starting_locations = configuration.starting_location.locations
    if len(starting_locations) == 1:
        area = game_description.world_list.area_by_area_location(next(iter(starting_locations)))
        starting_location = game_description.world_list.area_name(area)
    else:
        starting_location = "{} locations".format(len(starting_locations))

    template_strings["Gameplay"].append(f"Starting Location: {starting_location}")

    return template_strings


def fill_template_strings_from_tree(template_strings: Dict[str, List[str]], tree: Dict[str, List[Dict[str, bool]]]):
    for category, entries in tree.items():
        if category not in template_strings:
            template_strings[category] = []

        for entry in entries:
            messages = [message for message, flag in entry.items() if flag]
            if messages:
                template_strings[category].append(", ".join(messages))


def _echoes_format_params(configuration: EchoesConfiguration) -> Dict[str, List[str]]:
    major_items = configuration.major_items_configuration
    item_database = default_database.item_database_for_game(configuration.game)

    template_strings = _format_params_base(configuration)
    unified_ammo = configuration.ammo_configuration.items_state[item_database.ammo["Beam Ammo Expansion"]]

    # Difficulty
    if (configuration.varia_suit_damage, configuration.dark_suit_damage) != (6, 1.2):
        template_strings["Difficulty"].append("Dark Aether: {:.2f} dmg/s Varia, {:.2f} dmg/s Dark".format(
            configuration.varia_suit_damage, configuration.dark_suit_damage
        ))

    if configuration.energy_per_tank != 100:
        template_strings["Difficulty"].append(f"Energy Tank: {configuration.energy_per_tank} energy")

    if configuration.safe_zone.heal_per_second != 1:
        template_strings["Difficulty"].append(f"Safe Zone: {configuration.safe_zone.heal_per_second:.2f} energy/s")

    extra_message_tree = {
        "Item Pool": [
            {
                "Progressive Suit": has_shuffled_item(major_items, "Progressive Suit"),
                "Progressive Grapple": has_shuffled_item(major_items, "Progressive Grapple"),
                "Split beam ammo": unified_ammo.pickup_count == 0,
            }
        ],
        "Difficulty": [
            {"1-HP Mode": configuration.dangerous_energy_tank},
        ],
        "Gameplay": [
            {f"Translator Gates: {configuration.translator_configuration.description()}": True},
            {f"Elevators: {configuration.elevators.description()}": not configuration.elevators.is_vanilla},
        ],
        "Game Changes": [
            message_for_required_mains(
                configuration.ammo_configuration,
                {
                    "Missiles needs Launcher": "Missile Expansion",
                    "Power Bomb needs Main": "Power Bomb Expansion",
                }
            ),
            {"Warp to start": configuration.warp_to_start,
             "Menu Mod": configuration.menu_mod,
             "Final bosses removed": configuration.elevators.skip_final_bosses},
        ]
    }
    fill_template_strings_from_tree(template_strings, extra_message_tree)

    # Sky Temple Keys
    if configuration.sky_temple_keys == LayoutSkyTempleKeyMode.ALL_BOSSES:
        template_strings["Item Pool"].append("Sky Temple Keys at all bosses")
    elif configuration.sky_temple_keys == LayoutSkyTempleKeyMode.ALL_GUARDIANS:
        template_strings["Item Pool"].append("Sky Temple Keys at all guardians")
    else:
        template_strings["Item Pool"].append(f"{configuration.sky_temple_keys.num_keys} Sky Temple Keys")

    return template_strings


def _corruption_format_params(configuration: CorruptionConfiguration) -> Dict[str, List[str]]:
    major_items = configuration.major_items_configuration
    template_strings = _format_params_base(configuration)

    extra_message_tree = {
        "Item Pool": [
            {
                "Progressive Missile": has_shuffled_item(major_items, "Progressive Missile"),
                "Progressive Beam": has_shuffled_item(major_items, "Progressive Beam"),
            }
        ],
        "Difficulty": [
            {f"Energy Tank: {configuration.energy_per_tank} energy": configuration.energy_per_tank != 100},
        ],
        "Gameplay": [
            {f"Teleporters: {configuration.elevators.description()}": not configuration.elevators.is_vanilla},
        ],
        "Game Changes": [
            message_for_required_mains(
                configuration.ammo_configuration,
                {
                    "Missiles needs Launcher": "Missile Expansion",
                    "Ship Missiles needs Main": "Ship Missile Expansion",
                }
            ),
            {"Final bosses removed": configuration.elevators.skip_final_bosses},
        ]
    }
    fill_template_strings_from_tree(template_strings, extra_message_tree)

    return template_strings


_PRIME1_CUTSCENE_MODE_DESCRIPTION = {
    LayoutCutsceneMode.MAJOR: "Major cutscene removal",
    LayoutCutsceneMode.MINOR: "Minor cutscene removal",
    LayoutCutsceneMode.COMPETITIVE: "Competitive cutscene removal",
    LayoutCutsceneMode.ORIGINAL: None,
}


def _prime_format_params(configuration: PrimeConfiguration) -> Dict[str, List[str]]:
    template_strings = _format_params_base(configuration)
    cutscene_removal = _PRIME1_CUTSCENE_MODE_DESCRIPTION[configuration.qol_cutscenes]

    extra_message_tree = {
        "Difficulty": [
            {"Heat Damage: {:.2f} dmg/s".format(configuration.heat_damage): configuration.heat_damage != 10.0},
            {f"Energy Tank: {configuration.energy_per_tank} energy": configuration.energy_per_tank != 100},
        ],
        "Gameplay": [
            {f"Elevators: {configuration.elevators.description()}": not configuration.elevators.is_vanilla},
            {"Underwater movement without Gravity allowed": configuration.allow_underwater_movement_without_gravity},
        ],
        "Quality of Life": [
            {
                "Fixes to game breaking bugs": configuration.qol_game_breaking,
                "Pickup scans": configuration.qol_pickup_scans,
            }
        ],
        "Game Changes": [
            message_for_required_mains(
                configuration.ammo_configuration,
                {
                    "Missiles needs Launcher": "Missile Expansion",
                    "Power Bomb needs Main": "Power Bomb Expansion",
                }
            ),
            {
                "Varia-only heat protection": configuration.heat_protection_only_varia,
                "Progressive suit damage reduction": configuration.progressive_damage_reduction,
            },
            {
                "Warp to start": configuration.warp_to_start,
                "Final bosses removed": configuration.elevators.skip_final_bosses,
                "Unlocked Vault door": configuration.main_plaza_door,
                "Phazon Elite without Dynamo": configuration.phazon_elite_without_dynamo,
            },
            {
                "Small Samus": configuration.small_samus,
            },
            {
                cutscene_removal: cutscene_removal is not None,
            }
        ],
    }

    fill_template_strings_from_tree(template_strings, extra_message_tree)

    backwards = [
        message
        for flag, message in [
            (configuration.backwards_frigate, "Frigate"),
            (configuration.backwards_labs, "Labs"),
            (configuration.backwards_upper_mines, "Upper Mines"),
            (configuration.backwards_lower_mines, "Lower Mines"),
        ]
        if flag
    ]
    if backwards:
        template_strings["Game Changes"].append("Allowed backwards: {}".format(", ".join(backwards)))

    # Artifacts
    template_strings["Item Pool"].append(f"{configuration.artifact_target.num_artifacts} Artifacts, "
                                         f"{configuration.artifact_minimum_progression} min actions")

    return template_strings


def describe(preset: Preset) -> Iterable[PresetDescription]:
    configuration = preset.configuration

    if preset.game == RandovaniaGame.METROID_PRIME_ECHOES:
        template_strings = _echoes_format_params(configuration)

    elif preset.game == RandovaniaGame.METROID_PRIME_CORRUPTION:
        template_strings = _corruption_format_params(configuration)

    elif preset.game == RandovaniaGame.METROID_PRIME:
        template_strings = _prime_format_params(configuration)

    else:
        template_strings = _format_params_base(configuration)

    if configuration.multi_pickup_placement:
        template_strings["Item Placement"].append("Multi-pickup placement")

    for category, entries in template_strings.items():
        if entries:
            yield category, entries


def merge_categories(categories: Iterable[PresetDescription]) -> str:
    return "".join(
        """<h4><span style="font-weight:600;">{}</span></h4><p>{}</p>""".format(category, "<br />".join(items))
        for category, items in categories
    )



from randovania.games.game import RandovaniaGame
from randovania.games.prime.all_prime_dol_patches import StringDisplayPatchAddresses, PowerupFunctionsAddresses
from randovania.games.prime.corruption_dol_patches import CorruptionDolVersion

ALL_VERSIONS = [
    CorruptionDolVersion(
        game=RandovaniaGame.METROID_PRIME_CORRUPTION,
        description="Wii NTSC",
        build_string_address=0x805822b0,
        build_string=b"!#$MetroidBuildInfo!#$2007/07/27 13:13 Build v3.436 (MP3)",
        sda2_base=0x806869C0,
        sda13_base=0x806801c0,
        game_state_pointer=0x8067dc0c,
        cplayer_vtable=0x80592c78,
        cstate_manager_global=0x805c4f70,

        string_display=None,
        # string_display=StringDisplayPatchAddresses(
        #     update_hint_state=None,
        #     message_receiver_string_ref=0x805a4200,
        #     wstring_constructor=None,
        #     display_hud_memo=0x801c6480,
        #     max_message_size=200,
        # ),
        powerup_functions=PowerupFunctionsAddresses(
            add_power_up=0x8019111c,
            incr_pickup=0x801913a0,
            decr_pickup=0x8019151c,
        ),
    ),
]



import copy
from typing import List, Any, Tuple, Optional

from PySide2.QtCore import QTimer, Signal, QObject
from qasync import asyncSlot

from randovania.game_connection.connection_backend import ConnectionBackend
from randovania.game_connection.connection_base import LocationListener
from randovania.game_connection.executor.memory_operation import MemoryOperationExecutor
from randovania.game_description.resources.pickup_entry import PickupEntry


class GameConnection(QObject, ConnectionBackend):
    Updated = Signal()

    _dt: float = 2.5
    _last_status: Any = None
    _permanent_pickups: List[Tuple[str, PickupEntry]]

    def __init__(self, executor: MemoryOperationExecutor):
        super().__init__()
        ConnectionBackend.__init__(self, executor)
        self._permanent_pickups = []

        self._timer = QTimer(self)
        self._timer.timeout.connect(self._auto_update)
        self._timer.setInterval(self._dt * 1000)
        self._timer.setSingleShot(True)
        self._notify_status()

    def set_executor(self, executor: MemoryOperationExecutor):
        self.executor = executor
        self._notify_status()

    async def start(self):
        self._timer.start()

    async def stop(self):
        self._timer.stop()

    @asyncSlot()
    async def _auto_update(self):
        try:
            await self.update(self._dt)
            self._notify_status()
        finally:
            self._timer.start()

    def _notify_status(self):
        new_status = self.current_status
        inventory = self.get_current_inventory()

        if self._last_status != (new_status, self.executor, inventory):
            self._last_status = (new_status, self.executor, copy.copy(inventory))
            self.Updated.emit()

    @property
    def pretty_current_status(self) -> str:
        return f"{self.backend_choice.pretty_text}: {self.current_status.pretty_text}"

    @property
    def current_game_name(self) -> Optional[str]:
        if self.connector is not None:
            return self.connector.game_enum.long_name

    @property
    def name(self) -> str:
        raise ValueError("bleh")

    def set_location_collected_listener(self, listener: Optional[LocationListener]):
        super(ConnectionBackend, self).set_location_collected_listener(listener)
        self.checking_for_collected_index = listener is not None




import re
from typing import List, Tuple

from PySide2.QtWidgets import QDialog, QWidget

from randovania.game_description.game_description import GameDescription
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_info import ResourceInfo
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.game_description.world.area import Area
from randovania.game_description.world.node import DockNode
from randovania.game_description.world.world import World
from randovania.gui.generated.trick_details_popup_ui import Ui_TrickDetailsPopup
from randovania.gui.lib.common_qt_lib import set_default_window_icon
from randovania.gui.lib.window_manager import WindowManager
from randovania.layout.base.trick_level import LayoutTrickLevel


def _requirement_at_value(resource: ResourceInfo, level: LayoutTrickLevel):
    return lambda individual: individual.resource == resource and individual.amount == level.as_number


def _area_uses_resource(area: Area,
                        criteria,
                        database: ResourceDatabase,
                        ) -> bool:
    """
    Checks the area RequirementSet in the given Area uses the given trick at the given level.
    :param area:
    :param resource:
    :param level:
    :return:
    """

    def _uses_trick(requirements: Requirement) -> bool:
        return any(criteria(individual) for individual in requirements.as_set(database).all_individual)

    for node in area.nodes:
        if isinstance(node, DockNode):
            if _uses_trick(node.default_dock_weakness.requirement):
                return True

        if any(_uses_trick(req) for req in area.connections[node].values()):
            return True

    return False


class BaseResourceDetailsPopup(QDialog, Ui_TrickDetailsPopup):
    def __init__(self,
                 parent: QWidget,
                 window_manager: WindowManager,
                 game_description: GameDescription,
                 areas_to_show: List[Tuple[World, Area]],
                 ):
        super().__init__(parent)
        self.setupUi(self)
        set_default_window_icon(self)

        self._window_manager = window_manager
        self._game_description = game_description

        # setup
        self.area_list_label.linkActivated.connect(self._on_click_link_to_data_editor)

        # connect
        self.button_box.accepted.connect(self.button_box_close)
        self.button_box.rejected.connect(self.button_box_close)

        # Update
        if areas_to_show:
            lines = [
                (f'<a href="data-editor://{world.correct_name(area.in_dark_aether)}/{area.name}">'
                 f'{world.correct_name(area.in_dark_aether)} - {area.name}</a>')
                for (world, area) in areas_to_show
            ]
            self.area_list_label.setText("<br />".join(sorted(lines)))
        else:
            self.area_list_label.setText("This trick is not used in this level.")

    def button_box_close(self):
        self.reject()

    def _on_click_link_to_data_editor(self, link: str):
        info = re.match(r"^data-editor://([^)]+)/([^)]+)$", link)
        if info:
            world_name, area_name = info.group(1, 2)
            self._window_manager.open_data_visualizer_at(world_name, area_name, game=self._game_description.game)


class TrickDetailsPopup(BaseResourceDetailsPopup):
    def __init__(self,
                 parent: QWidget,
                 window_manager: WindowManager,
                 game_description: GameDescription,
                 trick: TrickResourceInfo,
                 level: LayoutTrickLevel,
                 ):
        areas_to_show = [
            (world, area)
            for world in game_description.world_list.worlds
            for area in world.areas
            if _area_uses_resource(area, _requirement_at_value(trick, level), game_description.resource_database)
        ]
        super().__init__(parent, window_manager, game_description, areas_to_show)

        # setup
        self.setWindowTitle(f"Trick Details: {trick.long_name} at {level.long_name}")
        self.title_label.setText(self.title_label.text().format(
            trick=trick,
            level=level.long_name,
        ))


class ResourceDetailsPopup(BaseResourceDetailsPopup):
    def __init__(self,
                 parent: QWidget,
                 window_manager: WindowManager,
                 game_description: GameDescription,
                 resource: ResourceInfo,
                 ):
        areas_to_show = [
            (world, area)
            for world in game_description.world_list.worlds
            for area in world.areas
            if _area_uses_resource(area,
                                   lambda individual: individual.resource == resource,
                                   game_description.resource_database)
        ]
        super().__init__(parent, window_manager, game_description, areas_to_show)

        # setup
        self.setWindowTitle(f"Details for {resource.long_name}")
        self.title_label.setText(f"""<html><head/><body>
        <p><span style=" font-weight:600;">{resource.long_name}</span></p>
        <p>{resource.long_name} can be found in the rooms listed below.</p>
        <p>Click a room name to open it in the Data Visualizer for more details.</p>
        </body></html>""")



from PySide2 import QtCore

from randovania.gui.dialog.echoes_cosmetic_patches_dialog import EchoesCosmeticPatchesDialog
from randovania.gui.dialog.prime_cosmetic_patches_dialog import PrimeCosmeticPatchesDialog
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches
from randovania.layout.prime2.echoes_user_preferences import EchoesUserPreferences, SoundMode


def test_qol_cosmetic(skip_qtbot):
    cosmetic_patches = PrimeCosmeticPatches(qol_cosmetic=False)

    dialog = PrimeCosmeticPatchesDialog(None, cosmetic_patches)
    skip_qtbot.addWidget(dialog)

    skip_qtbot.mouseClick(dialog.qol_cosmetic_check, QtCore.Qt.LeftButton)

    assert dialog.cosmetic_patches == PrimeCosmeticPatches(qol_cosmetic=True)


def test_open_map(skip_qtbot):
    cosmetic_patches = PrimeCosmeticPatches(open_map=True)

    dialog = PrimeCosmeticPatchesDialog(None, cosmetic_patches)
    skip_qtbot.addWidget(dialog)

    skip_qtbot.mouseClick(dialog.open_map_check, QtCore.Qt.LeftButton)

    assert dialog.cosmetic_patches == PrimeCosmeticPatches(open_map=False)



from PySide2 import QtCore

from randovania.gui.dialog.corruption_cosmetic_patches_dialog import CorruptionCosmeticPatchesDialog
from randovania.gui.dialog.echoes_cosmetic_patches_dialog import EchoesCosmeticPatchesDialog
from randovania.gui.dialog.prime_cosmetic_patches_dialog import PrimeCosmeticPatchesDialog
from randovania.layout.prime1.prime_cosmetic_patches import PrimeCosmeticPatches
from randovania.layout.prime2.echoes_cosmetic_patches import EchoesCosmeticPatches
from randovania.layout.prime2.echoes_user_preferences import EchoesUserPreferences, SoundMode
from randovania.layout.prime3.corruption_cosmetic_patches import CorruptionCosmeticPatches, CorruptionSuit


def test_random_welding_colors(skip_qtbot):
    cosmetic_patches = CorruptionCosmeticPatches(random_welding_colors=False)

    dialog = CorruptionCosmeticPatchesDialog(None, cosmetic_patches)
    skip_qtbot.addWidget(dialog)

    skip_qtbot.mouseClick(dialog.random_welding_colors_check, QtCore.Qt.LeftButton)

    assert dialog.cosmetic_patches == CorruptionCosmeticPatches(random_welding_colors=True)


def test_open_map(skip_qtbot):
    cosmetic_patches = CorruptionCosmeticPatches(player_suit=CorruptionSuit.CORRUPTED_25)

    dialog = CorruptionCosmeticPatchesDialog(None, cosmetic_patches)
    skip_qtbot.addWidget(dialog)

    skip_qtbot.keyClicks(dialog.suit_combo, "PED")

    assert dialog.cosmetic_patches == CorruptionCosmeticPatches(player_suit=CorruptionSuit.PED)



import collections

import typing
from typing import Tuple, Optional

from randovania.game_description.item.item_category import USELESS_ITEM_CATEGORY, ItemCategory
from randovania.game_description.assignment import PickupAssignment, PickupTarget
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import HintItemPrecision, Hint, RelativeDataItem
from randovania.game_description.resources.pickup_entry import PickupEntry, PickupModel
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.world.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.games.prime.patcher_file_lib import hint_lib
from randovania.games.prime.patcher_file_lib.hint_formatters import RelativeFormatter
from randovania.games.prime.patcher_file_lib.hint_lib import Determiner
from randovania.interface_common.players_configuration import PlayersConfiguration

_DET_AN = [
    "Annihilator Beam",
    "Amber Translator",
    "Echo Visor",
    "Emerald Translator",
    "Energy Transfer Module",
    "Energy Tank",
    "Ice Beam",
    "Ice Spreader",
    "X-Ray Visor"
]

_DET_NULL = []
_DET_NULL.extend(f"{temple} Key {i}"
                 for i in range(1, 4)
                 for temple in ("Dark Agon", "Dark Torvus", "Ing Hive"))
_DET_NULL.extend(f"Sky Temple Key {i}" for i in range(1, 10))


class RelativeItemFormatter(RelativeFormatter):
    def __init__(self, world_list: WorldList, patches: GamePatches, players_config: PlayersConfiguration):
        super().__init__(world_list, patches)
        self.players_config = players_config

    def format(self, determiner: Determiner, pickup: str, hint: Hint) -> str:
        relative = typing.cast(RelativeDataItem, hint.precision.relative)
        index = relative.other_index

        other_area = self.world_list.nodes_to_area(self._index_to_node[index])
        other_determiner, name = create_pickup_hint(self.patches.pickup_assignment, self.world_list,
                                                    relative.precision,
                                                    self.patches.pickup_assignment.get(index),
                                                    self.players_config,
                                                    False)
        other_name = f"{other_determiner}{name}"

        return self.relative_format(determiner, pickup, hint, other_area, other_name)


def _calculate_determiner(pickup_assignment: PickupAssignment, pickup: PickupEntry, world_list: WorldList) -> str:
    name_count = collections.defaultdict(int)
    for i in range(world_list.num_pickup_nodes):
        index = PickupIndex(i)
        if index in pickup_assignment:
            pickup_name = pickup_assignment[index].pickup.name
        else:
            pickup_name = "Energy Transfer Module"
        name_count[pickup_name] += 1

    if pickup.name in _DET_NULL:
        determiner = ""
    elif name_count[pickup.name] == 1:
        determiner = "the "
    elif pickup.name in _DET_AN:
        determiner = "an "
    else:
        determiner = "a "

    return determiner


def create_pickup_hint(pickup_assignment: PickupAssignment,
                       world_list: WorldList,
                       precision: HintItemPrecision,
                       target: Optional[PickupTarget],
                       players_config: PlayersConfiguration,
                       include_owner: bool,
                       ) -> Tuple[Determiner, str]:
    """

    :param pickup_assignment:
    :param world_list:
    :param precision:
    :param target:
    :param players_config:
    :param include_owner:
    :return:
    """
    if target is None:
        target = PickupTarget(
            pickup=PickupEntry(
                name="Energy Transfer Module",
                progression=tuple(),
                model=PickupModel(
                    game=RandovaniaGame.METROID_PRIME_ECHOES,
                    name="EnergyTransferModule",
                ),
                item_category=USELESS_ITEM_CATEGORY,
                broad_category=USELESS_ITEM_CATEGORY,
            ),
            player=players_config.player_index,
        )

    if precision is HintItemPrecision.GENERAL_CATEGORY:
        details = target.pickup.item_category.general_details

    elif precision is HintItemPrecision.PRECISE_CATEGORY:
        details = target.pickup.item_category.hint_details

    elif precision is HintItemPrecision.BROAD_CATEGORY:
        details = target.pickup.broad_category.hint_details

    elif precision is HintItemPrecision.DETAILED:
        details = _calculate_determiner(pickup_assignment, target.pickup, world_list), target.pickup.name

    elif precision is HintItemPrecision.NOTHING:
        details = "an ", "item"

    else:
        raise ValueError(f"Unknown precision: {precision}")

    if include_owner and players_config.is_multiworld:
        determiner = hint_lib.player_determiner(players_config, target.player)
    else:
        determiner = Determiner(details[0], True)

    return determiner, details[1]



from typing import Dict

from randovania.game_description.game_patches import GamePatches
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.games.prime import echoes_items
from randovania.games.prime.patcher_file_lib import hint_lib, guaranteed_item_hint
from randovania.interface_common.players_configuration import PlayersConfiguration

_SKY_TEMPLE_KEY_SCAN_ASSETS = [
    0xD97685FE,
    0x32413EFD,
    0xDD8355C3,
    0x3F5F4EBA,
    0xD09D2584,
    0x3BAA9E87,
    0xD468F5B9,
    0x2563AE34,
    0xCAA1C50A,
]


def _sky_temple_key_name(key_number: int) -> str:
    return hint_lib.color_text(hint_lib.TextColor.ITEM, f"Sky Temple Key {key_number}")


def create_hints(all_patches: Dict[int, GamePatches],
                 players_config: PlayersConfiguration,
                 resource_database: ResourceDatabase,
                 area_namers: Dict[int, hint_lib.AreaNamer],
                 hide_area: bool,
                 ) -> list:
    """
    Creates the string patches entries that changes the Sky Temple Gateway hint scans with hints for where
    the STK actually are.
    :param all_patches:
    :param players_config:
    :param resource_database:
    :param area_namers:
    :param hide_area: Should the hint include only the world?
    :return:
    """
    resulting_hints = guaranteed_item_hint.create_guaranteed_hints_for_resources(
        all_patches, players_config, area_namers, hide_area,
        [resource_database.get_item(index) for index in echoes_items.SKY_TEMPLE_KEY_ITEMS],
        hint_lib.TextColor.ITEM,
        hint_lib.TextColor.LOCATION,
    )
    return [
        hint_lib.create_simple_logbook_hint(
            _SKY_TEMPLE_KEY_SCAN_ASSETS[key_number],
            resulting_hints[resource_database.get_item(key_index)],
        )
        for key_number, key_index in enumerate(echoes_items.SKY_TEMPLE_KEY_ITEMS)
    ]


def hide_hints() -> list:
    """
    Creates the string patches entries that changes the Sky Temple Gateway hint scans with hints for
    completely useless text.
    :return:
    """

    return [
        hint_lib.create_simple_logbook_hint(
            _SKY_TEMPLE_KEY_SCAN_ASSETS[key_number],
            "{} is lost somewhere in Aether.".format(_sky_temple_key_name(key_number + 1)),
        )
        for key_number in range(9)
    ]



import asyncio

import pytest
from mock import MagicMock, AsyncMock, call

from randovania.game_connection.executor.memory_operation import MemoryOperationException, MemoryOperation
from randovania.game_connection.executor.nintendont_executor import NintendontExecutor, SocketHolder, RequestBatch


@pytest.fixture(name="executor")
def nintendont_executor():
    executor = NintendontExecutor("localhost")
    return executor


@pytest.mark.asyncio
async def test_perform_memory_operations_success(executor: NintendontExecutor):
    executor._socket = MagicMock()
    executor._socket.max_input = 120
    executor._socket.max_output = 100
    executor._socket.max_addresses = 8
    executor._socket.writer.drain = AsyncMock()
    executor._socket.reader.read = AsyncMock(side_effect=[
        b"\x03" + b"A" * 50 + b"B" * 30,
        b"\x01" + b"C" * 60,
    ])
    ops = {
        MemoryOperation(0x1000, read_byte_count=50): b"A" * 50,
        MemoryOperation(0x1000, offset=10, read_byte_count=30, write_bytes=b"1" * 30): b"B" * 30,
        MemoryOperation(0x1000, read_byte_count=60): b"C" * 60,
    }

    # Run
    result = await executor.perform_memory_operations(list(ops.keys()))

    # Assert
    executor._socket.writer.drain.assert_has_awaits([call(), call()])
    executor._socket.writer.write.assert_has_calls([
        call(b'\x00\x02\x01\x01\x00\x00\x10\x00' + b'\x80\x32' + b'\xd0\x1e\x00\n' + (b"1" * 30)),
        call(b'\x00\x01\x01\x01\x00\x00\x10\x00\x80\x3c'),
    ])
    assert result == ops
    executor._socket.reader.read.assert_has_awaits([call(1024), call(1024)])


@pytest.mark.asyncio
async def test_perform_memory_operations_invalid(executor: NintendontExecutor):
    executor._socket = MagicMock()
    executor._socket.max_input = 120
    executor._socket.max_output = 100
    executor._socket.max_addresses = 8
    executor._socket.writer.drain = AsyncMock()
    executor._socket.reader.read = AsyncMock(side_effect=[
        b"\x03" + b"A" * 50 + b"B" * 30,
    ])

    # Run
    with pytest.raises(MemoryOperationException):
        await executor.perform_memory_operations([
            MemoryOperation(0x1000, read_byte_count=50),
            MemoryOperation(0x2000, read_byte_count=10),
            MemoryOperation(0x2000, read_byte_count=10),
        ])

    # Assert
    executor._socket.writer.drain.assert_has_awaits([call()])
    executor._socket.writer.write.assert_has_calls([
        call(b'\x00\x03\x02\x01\x00\x00\x10\x00\x00\x00 \x00\x802\x81\n\x81\n'),
    ])
    executor._socket.reader.read.assert_has_awaits([call(1024)])


@pytest.mark.asyncio
async def test_perform_single_giant_memory_operation(executor: NintendontExecutor):
    executor._socket = MagicMock()
    executor._socket.max_input = 120
    executor._socket.max_output = 100
    executor._socket.max_addresses = 8
    executor._socket.writer.drain = AsyncMock()
    executor._socket.reader.read = AsyncMock(side_effect=[
        b"\x01",
        b"\x01",
    ])

    # Run
    result = await executor.perform_single_memory_operation(
        MemoryOperation(0x1000, write_bytes=b"1" * 200),
    )

    # Assert
    executor._socket.writer.drain.assert_has_awaits([call(), call()])
    executor._socket.writer.write.assert_has_calls([
        call(b'\x00\x01\x01\x01\x00\x00\x10\x00' + b'\x40\x64' + (b"1" * 100)),
        call(b'\x00\x01\x01\x01\x00\x00\x10\x64' + b'\x40\x64' + (b"1" * 100)),
    ])
    assert result is None
    executor._socket.reader.read.assert_has_awaits([call(1024), call(1024)])


@pytest.mark.asyncio
async def test_connect(executor, mocker):
    reader, writer = MagicMock(), MagicMock()
    writer.drain = AsyncMock()
    reader.read = AsyncMock(return_value=b'\x00\x00\x00\x02\x00\x00\x00x\x00\x00\x00\xfa\x00\x00\x00\x03')
    mock_open = mocker.patch("asyncio.open_connection", new_callable=AsyncMock, return_value=(reader, writer))

    # Run
    await executor.connect()

    # Assert
    mock_open.assert_awaited_once_with("localhost", executor._port)
    writer.drain.assert_awaited_once_with()

    socket: SocketHolder = executor._socket
    assert socket.reader is reader
    assert socket.writer is writer
    assert socket.api_version == 2
    assert socket.max_input == 120
    assert socket.max_output == 250
    assert socket.max_addresses == 3


@pytest.mark.asyncio
async def test_connect_when_connected(executor: NintendontExecutor):
    executor._socket = True
    assert await executor.connect()


@pytest.mark.asyncio
async def test_connect_invalid_ip(executor: NintendontExecutor):
    # Setup
    executor._ip = "127..0.0.1"

    # Run
    assert not await executor.connect()

    # Assert
    assert executor._socket is None
    assert type(executor._socket_error) is UnicodeError
    assert "encoding with 'idna' codec failed" in str(executor._socket_error)


@pytest.mark.asyncio
async def test_disconnect_not_connected(executor: NintendontExecutor):
    await executor.disconnect()
    assert executor._socket is None


@pytest.mark.asyncio
async def test_disconnect_connected(executor: NintendontExecutor):
    # Setup
    socket = MagicMock()
    executor._socket = socket

    # Run
    await executor.disconnect()

    # Assert
    assert executor._socket is None
    socket.writer.close.assert_called_once_with()


@pytest.mark.parametrize("use_timeout", [False, True])
@pytest.mark.asyncio
async def test_send_requests_to_socket_timeout(executor: NintendontExecutor, use_timeout):
    # Setup
    socket = AsyncMock()
    socket.writer.drain.side_effect = asyncio.TimeoutError() if use_timeout else OSError("test-exp")
    executor._socket = socket
    executor.disconnect = AsyncMock()
    reqs = [RequestBatch()]
    if use_timeout:
        msg = "Timeout when reading response"
    else:
        msg = "Unable to send 1 requests: test-exp"

    # Run
    with pytest.raises(MemoryOperationException, match=msg) as x:
        await executor._send_requests_to_socket(reqs)

    # Assert
    assert executor._socket_error is x.value
    executor.disconnect.assert_awaited_once_with()




from enum import Enum

from randovania.bitpacking.bitpacking import BitPackEnum


class LayoutArtifactMode(BitPackEnum, Enum):
    ZERO = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9
    TEN = 10
    ELEVEN = 11
    TWELVE = 12

    @property
    def num_artifacts(self):
        return self.value



import copy
import itertools
from typing import FrozenSet, List, Tuple, Set

from randovania.game_description import game_description
from randovania.game_description.world.node import ResourceNode
from randovania.game_description.requirements import RequirementList, ResourceRequirement, RequirementSet
from randovania.game_description.resources import resource_info
from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_info import CurrentResources, ResourceInfo
from randovania.game_description.resources.resource_type import ResourceType
from randovania.generator.generator_reach import GeneratorReach
from randovania.resolver import debug
from randovania.resolver.state import State

PickupCombination = Tuple[PickupEntry, ...]
PickupCombinations = Tuple[PickupCombination, ...]


def _resources_in_pickup(pickup: PickupEntry, current_resources: CurrentResources) -> FrozenSet[ResourceInfo]:
    resource_gain = pickup.resource_gain(current_resources, force_lock=True)
    return frozenset(resource for resource, _ in resource_gain)


def interesting_resources_for_reach(reach: GeneratorReach) -> FrozenSet[ResourceInfo]:
    satisfiable_requirements: FrozenSet[RequirementList] = frozenset(itertools.chain.from_iterable(
        requirements.alternatives
        for requirements in reach.unreachable_nodes_with_requirements().values()
    ))
    return game_description.calculate_interesting_resources(
        satisfiable_requirements,
        reach.state.resources,
        reach.state.energy,
        reach.state.resource_database
    )


def _unsatisfied_item_requirements_in_list(alternative: RequirementList,
                                           state: State,
                                           uncollected_resources: List[ResourceInfo]):
    possible = True
    items = []
    damage = []

    for individual in alternative.items:
        if individual.negate:
            possible = False
            break

        if individual.resource.resource_type == ResourceType.DAMAGE:
            damage.append(individual)
            continue

        if individual.satisfied(state.resources, state.energy, state.resource_database):
            continue

        if individual.resource.resource_type != ResourceType.ITEM:
            if individual.resource not in uncollected_resources:
                possible = False
                break

        items.append(individual)

    if not possible:
        return

    sum_damage = sum(req.damage(state.resources, state.resource_database) for req in damage)
    if state.energy < sum_damage:
        tank_count = (sum_damage - state.energy) // state.game_data.energy_per_tank
        yield items + [ResourceRequirement(state.resource_database.energy_tank, tank_count + 1, False)]
        # FIXME: get the required items for reductions (aka suits)
    else:
        yield items


def _requirement_lists_without_satisfied_resources(state: State,
                                                   possible_sets: List[RequirementSet],
                                                   uncollected_resources: List[ResourceInfo],
                                                   ) -> Set[RequirementList]:
    seen_lists = set()
    result = set()

    def _add_items(it):
        items_tuple = RequirementList(it)
        if items_tuple not in result:
            result.add(items_tuple)

    for requirements in possible_sets:
        # Maybe should first recreate `requirements` by removing the satisfied items or the ones that can't be
        for alternative in requirements.alternatives:
            if alternative in seen_lists:
                continue
            seen_lists.add(alternative)

            for items in _unsatisfied_item_requirements_in_list(alternative, state, uncollected_resources):
                _add_items(items)

    if debug.debug_level() > 2:
        print(">> All requirement lists:")
        for items in sorted(result):
            print(f"* {items}")

    return result


def pickups_to_solve_list(pickup_pool: List[PickupEntry],
                          requirement_list: RequirementList,
                          state: State):
    pickups = []

    resources = copy.copy(state.resources)
    pickups_for_this = list(pickup_pool)

    for individual in sorted(requirement_list.values()):
        if individual.satisfied(resources, state.energy, state.resource_database):
            continue

        # FIXME: this picks Dark Beam to provide Dark Ammo

        # Create another copy of the list so we can remove elements while iterating
        for pickup in list(pickups_for_this):
            new_resources = resource_info.convert_resource_gain_to_current_resources(
                pickup.resource_gain(resources))
            pickup_progression = resource_info.convert_resource_gain_to_current_resources(pickup.progression)
            if new_resources.get(individual.resource, 0) + pickup_progression.get(individual.resource, 0) > 0:
                pickups.append(pickup)
                pickups_for_this.remove(pickup)
                resource_info.add_resources_into_another(resources, new_resources)

            if individual.satisfied(resources, state.energy, state.resource_database):
                break

        if not individual.satisfied(resources, state.energy, state.resource_database):
            return None

    return pickups


def get_pickups_that_solves_unreachable(pickups_left: List[PickupEntry],
                                        reach: GeneratorReach,
                                        uncollected_resource_nodes: List[ResourceNode],
                                        ) -> PickupCombinations:
    """New logic. Given pickup list and a reach, checks the combination of pickups
    that satisfies on unreachable nodes"""
    state = reach.state
    possible_sets = list(reach.unreachable_nodes_with_requirements().values())
    uncollected_resources = [node.resource() for node in uncollected_resource_nodes]

    all_lists = _requirement_lists_without_satisfied_resources(state, possible_sets, uncollected_resources)

    result = []
    for requirement_list in sorted(all_lists, key=lambda it: it.as_stable_sort_tuple):
        pickups = pickups_to_solve_list(pickups_left, requirement_list, state)
        if pickups is not None and pickups:
            # FIXME: avoid duplicates in result
            result.append(tuple(pickups))

    if debug.debug_level() > 2:
        print(">> All pickup combinations alternatives:")
        for items in sorted(result):
            print("* {}".format(", ".join(p.name for p in items)))

    return tuple(result)


def get_pickups_with_interesting_resources(pickup_pool: List[PickupEntry],
                                           reach: GeneratorReach,
                                           uncollected_resource_nodes: List[ResourceNode],
                                           ) -> PickupCombinations:
    """Old logic. Given pickup list and a reach, gets these that gives at least one of the interesting resources."""
    interesting_resources = interesting_resources_for_reach(reach)
    progression_pickups = []

    for pickup in pickup_pool:
        if pickup in progression_pickups:
            continue
        if _resources_in_pickup(pickup, reach.state.resources).intersection(interesting_resources):
            progression_pickups.append(pickup)

    return tuple(
        (pickup,)
        for pickup in progression_pickups
    )
